"""
The MIT License (MIT)
Copyright (c) 2014-2015 University of Bristol

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
OR OTHER DEALINGS IN THE SOFTWARE.
"""
from datetime import timedelta, datetime
import xml.etree.ElementTree as ElementTree
import os
from datadiff import diff
import pytz
from mongoengine import connect, EmbeddedDocument
from mongoengine import Document, DateTimeField, StringField, EmbeddedDocumentListField, IntField
import logging
from collections import defaultdict

from sensor_base import SensorBase
from code_timer import CodeTimer
from all_annotations import ALL_ANNOTATIONS, CORRECTIONS


class HistoryItem(EmbeddedDocument):
    action = StringField(choices=('insert', 'delete'), required=True)
    key = StringField(required=True)
    value = StringField(required=True)


class HistoryField(EmbeddedDocument):
    datetime = DateTimeField(required=True)
    revision = IntField(min_value=0, required=True)
    diffs = EmbeddedDocumentListField(HistoryItem, required=True)


class Annotation(Document):
    start = DateTimeField(required=True)
    end = DateTimeField(required=True)
    label = StringField(db_field='label', required=True)
    # annotator = StringField(required=True)
    annotator = IntField(required=True)
    tier = StringField(db_field='tier', required=True)
    history = EmbeddedDocumentListField(HistoryField)

    meta = {
        'collection': 'annotations',
        'indexes': [{'fields': ['start']}],
        'ordering': ['start']
    }


class Annotations(SensorBase):
    def __init__(self, server_config, db, annotation_path):
        super(Annotations, self).__init__(server_config, db, "annotations")
        self.all_labels = defaultdict(set)
        # self.session = Session(db)
        self.session = connect(**dict((k, v) for k, v in server_config.items() if k not in ['modalities', 'summaries']))
        self.experiment_config = None
        self.annotation_path = annotation_path

    def init_experiment(self, config=None):
        if config:
            self.experiment_config = config

        if not self.experiment_config.annotations:
            logging.warn('No annotations defined for experiment')
            return

        self.docs = []
        for annotator in self.experiment_config.annotations:
            try:
                # docs = list(self.session.query(Annotation).filter(
                #     Annotation.start >= self.experiment_start,
                #     Annotation.end <= self.experiment_end,
                #     Annotation.annotator == annotator
                # ).ascending(Annotation.start).all())
                docs = list(Annotation.objects(
                    start__gte=self.experiment_config.experiment_start,
                    end__lte=self.experiment_config.experiment_end,
                    # annotator=annotator
                    annotator=annotator
                ))
            except TypeError as e:
                logging.warn(e)
                continue

            while len(docs) == 0:
                logging.info("No annotations found, re-importing")
                # import_annotations(self.collection, config)
                docs = import_annotation(annotator, self.collection, self.experiment_config, self.annotation_path)

            self.docs.extend(docs)

        logging.info("{0: <31}: {1} records found".format('Annotations:', len(self.docs)))

    def get_data(self):
        with CodeTimer('pulling annotation data from mongodb ...'):
            self.data = []
            for doc in self.docs:
                if not doc.label:
                    continue
                self.all_labels[doc.tier].add(doc.label)
                self.data.append({
                    'datetime': doc.start, 'tier': doc.tier, 'label': doc.label, 'trigger': 1, '_id': doc.id,
                    'annotator': doc.annotator})
                self.data.append({
                    'datetime': doc.end, 'tier': doc.tier, 'label': doc.label, 'trigger': -1, '_id': doc.id,
                    'annotator': doc.annotator})

            return self.data

    def get_tier(self, tier, annotator):
        if not self.data:
            self.data = self.get_data()
        return [d for d in self.data if d['tier'] == tier and d['annotator'] == int(annotator)]

    def get_annotator(self, annotator):
        if not self.data:
            self.data = self.get_data()
        return [d for d in self.data if d['annotator'] == int(annotator)]

    def post_process(self, df):
        for tier in self.all_labels:
            for annotator in self.experiment_config.annotations:
                labels = df[['label', 'trigger']][(df['tier'] == tier) & (df['annotator'] == float(annotator))]

                key = 'ann-' + annotator + '-' + tier + '-discrete'
                df[key] = -1

                for offset in xrange(0, len(labels), 2):
                    indices = (df.index > labels.index[offset]) & (df.index < labels.index[offset + 1])
                    label = labels.label.iloc[offset]

                    if not label or label == 'Empty':
                        continue

                    try:
                        if label in CORRECTIONS:
                            label = CORRECTIONS[label]
                        df.loc[indices, key] = ALL_ANNOTATIONS[tier].index(label)
                    except ValueError as e:
                        logging.debug(e)
                        # logging.debug(tier, label)

    def update(self, doc):
        # First make sure the document exists
        # doc = self.session.query(Annotation).filter(Annotation.mongo_id==doc['_id']).first()
        mongo_doc = Annotation.objects(id=doc['_id']).first()
        if not mongo_doc:
            return {'success': False, 'reason': 'id not found', '_id': mongo_doc.id, 'details': None}

        # if 'history' not in doc:
        #     doc['history'] = []

        # Check for differences between the document
        doc_no_history = dict(mongo_doc.to_mongo())
        doc_no_history.pop('history')
        doc.pop('_id')

        d = diff(doc_no_history, doc)
        diffs = [HistoryItem(action=x[0], key=y[0], value=y[1]) for x in d.diffs
                 if x[0] != 'equal' for y in x[1] if y[0] != '_id']

        if len(diffs) == 0:
            return {'success': False, 'reason': 'documents identical', '_id': mongo_doc.id, 'details': None}

        # We have diffs. Store them in the doc history
        # revision = len(doc['history'])
        # doc['history'].append({
        #     'revision': revision,
        #     'diffs': diffs,
        #     'datetime': pytz.UTC.localize(datetime.now())
        # })
        revision = len(mongo_doc.history)
        mongo_doc.history.append(
            HistoryField(revision=revision,
                         diffs=diffs,
                         datetime=pytz.UTC.localize(datetime.now()))
        )

        # update the document
        # for k in doc:
        #     mongo_doc[k] = doc[k]
        mongo_doc.update(**doc)

        # put into database
        # result = self.collection.update({'_id': doc['_id']}, doc, upsert=False, multi=False)
        mongo_doc.save()

        # if result['ok'] != 1 or result['n'] != 1 or not result['updatedExisting'] or result['nModified'] != 1:
        #     return {'success': False, 'reason': 'update failed', 'details': result, '_id': doc['_id']}

        # Finally update in-memory copy
        self.init_experiment()

        # TODO: Experiment should be updated now!
        return {'success': True, 'revision': revision, '_id': mongo_doc.id, 'diffs': diffs}


# def import_annotations(collection, experiment_config, annotation_path):
#     for annotator in experiment_config.annotations:
#         import_annotation(annotator, collection, experiment_config, annotation_path)


def import_annotation(annotator, collection, experiment_config, annotation_path):
    logging.info("Annotator: " + annotator)
    logging.info("First annotation: " + str(experiment_config.experiment_start))
    logging.info("Last annotation:  " + str(experiment_config.experiment_end))

    annotation_config = experiment_config.annotations[annotator]

    # remove old annotations from db if present
    try:
        result = collection.delete_many({
            'start': {'$gte': experiment_config.experiment_start},
            'end': {'$lte': experiment_config.experiment_end},
            'annotator': {'$eq': annotator}
        })
        count = result.deleted_count
    except TypeError:
        # For older versions of pymongo
        query = {
            'start': {'$gte': experiment_config.experiment_start},
            'end': {'$lte': experiment_config.experiment_end},
            'annotator': {'$eq': annotator}
        }
        count = collection.find(query).count()
        collection.remove(query)
    logging.info('{0} old annotations removed'.format(count))

    # Load in XML file
    tree = ElementTree.parse(os.path.join(annotation_path, annotation_config.filename))
    root = tree.getroot()
    # Create time slots as dictionary
    slots = dict((slot.attrib['TIME_SLOT_ID'], float(slot.attrib['TIME_VALUE']))
                 for slot in root.findall('./TIME_ORDER/TIME_SLOT'))
    tiers = []
    annotations = []
    all_labels = {}
    for tt in root.findall('./TIER'):
        tier = tt.attrib['TIER_ID']
        tiers.append(tier)
        all_labels[tier] = set()
        # print(tiers[-1])
        for ann in tt.findall('ANNOTATION/ALIGNABLE_ANNOTATION'):
            # print(ann.attrib)
            label = ann.find('ANNOTATION_VALUE').text

            # TODO: This is where the labels are missing
            if label is None or len(label) == 0:
                label = 'Empty'

            # Remove leading and trailing whitespace, and remove duplicate spaces (i.e. "  ")
            label = " ".join(label.split())

            if label in CORRECTIONS:
                label = CORRECTIONS[label]

            s1 = label[0:len(label) / 2].strip()
            if label[len(label) - len(s1):] == s1:
                label = s1

            label = label.strip()

            all_labels[tier].add(label)

            start = experiment_config.experiment_start + \
                timedelta(milliseconds=slots[ann.attrib['TIME_SLOT_REF1']]) + annotation_config.offset

            end = experiment_config.experiment_start + \
                timedelta(milliseconds=slots[ann.attrib['TIME_SLOT_REF2']]) + annotation_config.offset

            # annotation = {
            #     'tier': tier,
            #     'label': label,
            #     'start': start,
            #     'end': end,
            #     'annotator': annotator
            # }
            #
            # try:
            #     inserted_id = collection.insert_one(annotation).inserted_id
            # except TypeError:
            #     # for older versions of pymongo
            #     inserted_id = collection.insert(annotation)

            annotation = Annotation(tier=tier, label=label, start=start, end=end, annotator=annotator)
            annotation.save()

            # annotation['_id'] = inserted_id
            annotations.append(annotation)
    # logging.info("First annotation: " + str(experiment.experiment_start + timedelta(milliseconds=slots['ts1'])))
    # logging.info("Last annotation:  " + str(experiment.experiment_start +
    # timedelta(milliseconds=max(slots.values()))))
    logging.info('{0} Annotations inserted'.format(len(annotations)))
    return annotations
