"""
The MIT License (MIT)
Copyright (c) 2014-2015 University of Bristol

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
OR OTHER DEALINGS IN THE SOFTWARE.
"""
import pymongo
from utils import get_value_safe
from datetime import datetime
from tqdm import tqdm
import pytz
import logging


class Client(object):
    def __init__(self, server_config):
        self.server_config = server_config
        
        if 'connection_string' in server_config:
            self.client = pymongo.MongoClient(server_config['connection_string'])
            self.db = self.client[server_config['db']]
        else:
            self.client = pymongo.MongoClient(
                server_config['host'], 
                server_config['port'],
                tz_aware=self.get_config_value('tz_aware', True))
            self.db = self.client[server_config['db']]
            self.db.authenticate(server_config['username'], server_config['password'])

        self.collections = dict((k, self.db[v['collection']]) for k, v in server_config['modalities'].iteritems())

        self.summaries = [self.db[x] for x in server_config['summaries']]

    def get_config_value(self, key, default=None):
        return get_value_safe(self.server_config, key, default)

    def get_summaries(self, oldest_date):
        summaries = []

        for summary in self.summaries:
            docs = summary.find()  # {'_id': {'$gt', oldest_date}}).sort('_id', pymongo.ASCENDING)
            for doc in tqdm(docs, total=summary.count()):
                if 'value' not in doc or 'modality' not in doc['value'] or 'frequency' not in doc['value']:
                    # print('invalid: ' + str(doc['_id']))
                    continue

                if not isinstance(doc['_id'], datetime):
                    # print('wrong id: ' + str(doc['_id']))
                    continue

                if doc['_id'] < pytz.UTC.localize(oldest_date):
                    logging.debug('skipping: ' + str(doc['_id']))
                    continue

                modality = doc['value']['modality']
                frequency = doc['value']['frequency']
                counts = doc['value']['count']

                for uid in counts:
                    if type(counts[uid]) in [int, float]:
                        summaries.append({
                            't': doc['_id'].strftime("%Y-%m-%d %H:%M:%S"),
                            'name': modality,
                            'category': frequency,
                            'uid': uid,
                            'value': counts[uid]
                        })
                    elif type(counts[uid]) in [str, unicode] and counts[uid].startswith('undefined'):
                        continue
                    else:
                        for aid in counts[uid]:
                            if aid == "undefined":
                                continue
                            summaries.append({
                                't': doc['_id'].strftime("%Y-%m-%d %H:%M:%S"),
                                'name': modality,
                                'category': frequency,
                                'aid': aid,
                                'uid': uid,
                                'value': counts[uid][aid]
                            })

        return summaries

    def get_map_reduce_status(self):
        return list(self.db['IncrementalMRCounters'].find())
