"""
The MIT License (MIT)
Copyright (c) 2014-2015 University of Bristol

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
OR OTHER DEALINGS IN THE SOFTWARE.
"""
import numpy as np
from sensor_base import SensorBase
from code_timer import CodeTimer
from tqdm import tqdm


SENSOR_MAPPINGS = {
    'TEMP': 'temperature',
    'HUM': 'humidity',
    'DC': 'door',
    'WHT': 'hot-water',
    'LT': 'light',
    'WCD': 'cold-water',
    'NS': 'noise',
    'DT': 'dust',
    'PIR': 'motion',
    'ELEC': 'electricity'
}


MOTION_SENSORS = [
    "motion-S1_B1",
    "motion-S1_B0",
    "motion-S1_B2",
    "motion-S1_BR",
    "motion-S1_H",
    "motion-S1_K",
    "motion-S1_L",
    "motion-S1_S",
    "motion-S1_WC"
]


class Environmental(SensorBase):
    def __init__(self, server_config, db):
        super(Environmental, self).__init__(server_config, db, "environmental")

    def get_unique_sensor_names(self):
        # Get unique sensor names
        s = {'ELEC'}
        s.update(MOTION_SENSORS)
        s.update(self.collection.find({'sn': {'$ne': 'ELEC'}}).distinct('e.n'))
        s.update(self.collection.find({'sn': {'$ne': 'ELEC'}}).distinct('m.s'))
        return s

    def get_data(self):
        """
        Get data from mongodb
        :return: data as list
        """
        with CodeTimer('pulling environmental data from mongodb ...'):
            if not self.data:
                self.sensors = set()
                self.data = []
                for doc in tqdm(self.docs, total=self.docs.count()):
                    event = 'e' if 'e' in doc else 'm'
                    element_name = 'n' if event == 'e' else 's'
                    uid = doc['uid']

                    row = {'datetime': doc[self.date_time_field], 'uid': uid}
                    for v in SENSOR_MAPPINGS.values():
                        row[v] = None

                    for element in doc[event]:
                        sensor = SENSOR_MAPPINGS[doc['sn']] if 'sn' in doc else SENSOR_MAPPINGS[element[element_name]]
                        sid = sensor + '-' + uid
                        row[sid] = float(element['v']) if element['v'] is not None else np.nan
                        self.sensors.add(sid)

                    self.data.append(row)

            return self.data
