"""
The MIT License (MIT)
Copyright (c) 2014-2015 University of Bristol

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
OR OTHER DEALINGS IN THE SOFTWARE.
"""
import datetime
import logging
import pytz


def incremental_map_reduce(
        map_f,
        reduce_f,
        finalize_f,
        db,
        source_table_name,
        target_table_name,
        source_queued_date_field_name,
        query=None,
        scope=None,
        counter_table_name="IncrementalMRCounters",
        counter_key=None,
        min_datetime=None,
        max_datetime=None,
        reset=False,
        force=False):
    """ This method performs an incremental map-reduce on any new data in 'source_table_name'
    into 'target_table_name'.  It can be run in a cron job, for instance, and on each execution will
    process only the new, unprocessed records.
    The set of data to be processed incrementally is determined non-invasively (meaning the source table is not
    written to) by using the queued_date field 'source_queued_date_field_name'. When a record is ready to be processed,
    simply set its queued_date (which should be indexed for efficiency). When incremental_map_reduce() is run, any docs
    with queued_dates between the counter in 'counter_key' and 'max_datetime' will be map/reduced.
    If reset is True, it will clear 'target_table_name' and do a map reduce across all records older
    than max_datetime.
    If unspecified/None, counter_key defaults to counter_table_name:LastMaxDatetime.
    :param map_f: mapping function
    :param reduce_f: reduction function
    :param finalize_f: finalizing function
    :param db: database
    :param source_table_name: source table name
    :param target_table_name: target table name
    :param source_queued_date_field_name:
    :param query: Mongo query
    :param scope: Additional data to send
    :param counter_table_name: Where the counter is stored
    :param counter_key: Key of the counter
    :param min_datetime: Oldest record to look at
    :param max_datetime: Newest record to look at
    :param reset: Clear 'target_table_name' and do map reduce across all records older than max_datetime
    :param force: Force running, even if locked from previous runs
    """

    now = datetime.datetime.now()
    if max_datetime is None:
        max_datetime = now

    if reset:
        logging.debug("Resetting, dropping table " + target_table_name)
        db.drop_collection(target_table_name)

    time_limits = {"$lt": max_datetime}
    if min_datetime and isinstance(min_datetime, datetime.  datetime):
        time_limits.update({"$gte": min_datetime})

    if counter_key is None:
        counter_key = target_table_name + ":LastMaxDatetime"

    # If we've run before, filter out anything that we've processed already.
    last_max_datetime = None

    last_max_datetime_record = db[counter_table_name].find_and_modify(
        {'_id': counter_key},
        {'$set': {'inprogress': True}, '$push': {'m': now}},
        upsert=True
    )

    if force or last_max_datetime_record is None or not 'inprogress' in last_max_datetime_record:
        # first time ever run, or forced to go ahead anyway
        pass
    else:
        if last_max_datetime_record['inprogress']:
            if last_max_datetime_record['m'][0] < pytz.UTC.localize(now - datetime.timedelta(hours=2)):
                # lock timed out, so go ahead...
                logging.error(
                    target_table_name + " lock is old.  Ignoring it, but something was broken that caused it to not be "
                                        "unlocked...")
            else:
                logging.warning(target_table_name + " mapreduce already in progress, skipping...")
                raise RuntimeError(
                    target_table_name + " locked since %s.  Skipping..." % last_max_datetime_record['m'][0])

    if not reset:
        if last_max_datetime_record is not None:
            try:
                last_max_datetime = last_max_datetime_record['value']
                time_limits['$gt'] = last_max_datetime
                logging.debug('~FR limit last_max_datetime = %s' % (last_max_datetime,))
            except KeyError:
                # This happened on staging.  i guess it crashed somehow
                # between the find_and_modify and the final update?
                logging.error("~FR no value on message!")

    if isinstance(query, dict):
        query.update({source_queued_date_field_name: time_limits})
    else:
        query = {source_queued_date_field_name: time_limits}

    ret = db[source_table_name].map_reduce(
        map_f,
        reduce_f,
        finalize=finalize_f,
        out={'reduce': target_table_name},
        query=query,
        scope=scope,
        full_response=True
    )

    num_processed = ret['counts']['input']

    # Update our counter so we remember for the next pass.
    already_processed_through = db[counter_table_name].update(
        {'_id': counter_key},
        {'$set': {'inprogress': False, 'value': max_datetime}, '$unset': {'m': 1}},
        upsert=False,
        multi=False)

    logging.debug("Processed %d completed surveys from %s through %s.\nmap_reduce details: %s" % (
        num_processed, last_max_datetime, max_datetime, ret))

    return ret
