"""
The MIT License (MIT)
Copyright (c) 2014-2015 University of Bristol

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import logging
import logging.handlers
import logging.config
import time
# import requesocks as requests
import pprint
import sys
import warnings
from functools import wraps

# from exceptions import ConfigException
from . import __version__


def get_value_safe(d, key, default=None):
    return d[key] if key in d else default


class UTCFormatter(logging.Formatter):
    converter = time.gmtime


# LOGGING = {
#     'version': 1,
#     'disable_existing_loggers': True,
#     'formatters': {
#         'utc:': {
#             '()': UTCFormatter,
#             'format': "%(asctime)s [%(levelname)-5.5s]  %(message)s"
#         }
#     },
#     'handlers': {
#         'file': {
#             'class': 'logging.handlers.RotatingFileHandler',
#             'level': 'DEBUG',
#             'formatter': 'utc',
#             'filename': 'sphere_connector.log',
#             'mode': 'a',
#             'maxBytes': 10485760,
#             'backupCount': 5
#         },
#         'console': {
#             'class': 'logging.StreamHandler',
#             'level': 'DEBUG',
#             'formatter': 'utc'
#         }
#     }
# }


def handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    logging.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))


def initialise_logger(path='logs', filename='sphere_connector'):
    log_formatter = logging.Formatter("%(asctime)s [%(levelname)-5.5s]  %(message)s")
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    if not filename.endswith('.log'):
        filename += '.log'
    file_handler = logging.FileHandler(os.path.join(path, filename))
    file_handler.setFormatter(log_formatter)
    root_logger.addHandler(file_handler)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_formatter)
    root_logger.addHandler(console_handler)
    #
    # stream_handler = logging.StreamHandler()
    # stream_handler.setFormatter(log_formatter)
    # memory_handler = logging.handlers.MemoryHandler(1024 * 10, root_logger.level, stream_handler)
    # root_logger.addHandler(memory_handler)

    # Capture warnings
    logging.captureWarnings(True)

    # Capture uncaught exceptions
    sys.excepthook = handle_exception

    # logging.config.dictConfig(LOGGING)
    logging.debug("Sphere connector version: " + __version__)
    # return memory_handler


def deprecated(func):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    :param func: Function to apply to
    """
    def new_func(*args, **kwargs):
        warnings.warn("Call to deprecated function {}.".format(func.__name__),
                      category=DeprecationWarning)
        return func(*args, **kwargs)

    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    new_func.__dict__.update(func.__dict__)
    return new_func


class countcalls(object):
    """Decorator that keeps track of the number of times a function is called."""
    __instances = {}

    def __init__(self, f):
        self.__f = f
        self.__numcalls = 0
        countcalls.__instances[f] = self

    def __call__(self, *args, **kwargs):
        self.__numcalls += 1
        return self.__f(*args, **kwargs)

    @staticmethod
    def count(f):
        """
        Return the number of times the function f was called
        :param f: Function to count
        """
        return countcalls.__instances[f].__numcalls

    @staticmethod
    def counts():
        """
        Return a dict of {function: # of calls} for all registered functions.
        """
        return dict([(f, countcalls.count(f)) for f in countcalls.__instances])


class log_event(object):
    """
    Logging decorator that allows you to log with a specific logger.
    """
    # Customize these messages
    ENTRY_MESSAGE = 'Entering {}'
    EXIT_MESSAGE = 'Exiting {}'

    def __call__(self, func):
        """
        Returns a wrapper that wraps func.
        The wrapper will log the entry and exit points of the function
        with logging.INFO level.
        """
        # set logger if it was not set earlier
        @wraps(func)
        def wrapper(*args, **kwargs):
            logging.debug(self.ENTRY_MESSAGE.format(func.__name__))
            f_result = func(*args, **kwargs)
            logging.debug(self.EXIT_MESSAGE.format(func.__name__))
            return f_result
        return wrapper


class Printable(object):
    def __str__(self):
        pp = pprint.PrettyPrinter(indent=4)
        return pp.pformat(self.__dict__)
